/*
 * feedbackd - dynamic feedback system for LVS
 * Copyright (C) 2002 Jeremy Kerr
 * 
 * This file is part of feedbackd.
 *
 *  feedbackd is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  feedbackd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with feedbackd; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file feedbackd-master.c
 */

#include "feedbackd-master.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <assert.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "necp_listener.h"
#include "log.h"
#include "scheduler.h"
#include "ipvs_interface.h"

static int configure(const char *configfile, int logging);
static void dump_config(void);
	
const struct option options[] = {
	{"conffile",    1, 0, 'c'},
	{"dont-fork",   0, 0, 'd'},
	{"dry-run",     0, 0, 'n'},
	{"test-config", 0, 0, 't'},
	{"help",        0, 0, 'h'},
	{"usage",       0, 0, 'u'},
	{"version",     0, 0, 'V'},
	{0,             0, 0, 0 }
};

int main(int argc, char **argv)
{
	char *conffile;
	int c, dofork = 1, testconfig = 0;

	conffile = FEEDBACKD_CONF;
	
	while ((c = getopt_long(argc, argv, "c:dntV?hu", options, NULL))
	       != EOF) {
		switch(c) {
		case 'c':
			conffile = optarg;
			break;
		case 'd':
			dofork = 0;
			break;
		case 't':
			testconfig = 1;
			break;
		case 'n':
			ipvsif_set_dryrun(-1);
			break;
		case 'V':
			printf("%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
			return 0;
		case '?':
		case 'h':
		case 'u':
		default:
			print_usage(argv[0]);
			return -1;
		}
	}

	log_printf(LOG_INFO, "%s starting", argv[0]);
	log_printf(LOG_INFO, "Using config file %s", conffile);

	list_init(&servers);
	list_init(&services);
	list_init(&virtual_services);

	if (configure(conffile, !testconfig)) {
		log_printf(LOG_CRIT, "Configuration failed. Exiting");
		return -1;
	}

	if (testconfig) {
		log_setlevel(LOG_VDEBUG);
		/* dump_config(); */
		log_printf(LOG_VDEBUG, "Configuration successful");
		return 0;
	}

	dump_config();

	/* start necp_listener */
	necp_listener_running = 1;

	if (dofork && fork())
		return 0;

	if (ipvsif_init()) {
		log_printf(LOG_CRIT, "IPVS initialisation failed. Exiting");
		return -1;
	}

	if (ipvsif_init_virtual_services()) {
		log_printf(LOG_CRIT, "IPVS virtual server init failed. "
				"Exiting");
		return -1;
	}

	necp_listen();

	log_stop();
	ipvsif_close();


	return 0;
}
struct virtual_service *parse_virtual_service(xmlNodePtr node)
{
	char *value, *tmpserv = 0;
	struct protoent *proto = 0;
	struct virtual_service *service;
	struct servent *serv;

	if (!(service = (struct virtual_service *)malloc(sizeof(*service)))) {
		perror("malloc");
		return NULL;
	}

	memset(service, 0, sizeof(*service));

	if ((value = xmlGetProp(node, (const xmlChar *)"name"))) {
		strncpy(service->name, value, MAX_SERVICE_NAME_LEN - 1);
	} else {
		log_printf(LOG_ERR, "No name attribute for virtual service");
		free(service);
		return NULL;
	}

	node = node->children;
	while (node) {
		if (xmlIsBlankNode(node)) {
			node = node->next;
			continue;
		}
		
		if (!xmlStrcmp(node->name, (const xmlChar*)"port")) {
			value = xmlNodeListGetString(node->doc,
					             node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No port given in <port> "
						"tag, virtual service %s",
						service->name);
				free(service);
				return NULL;
			}

			tmpserv = value;
			
		} else if (!xmlStrcmp(node->name,
		                     (const xmlChar *)"protocol")) {
			value = xmlNodeListGetString(node->doc,
			                             node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No protocol given in "
						"<protocol> tag, virtual "
						"service %s",
						service->name);
				free(service);
			}

			if (!(proto = getprotobyname(value))) {
				log_printf(LOG_ERR, "Invalid protocol (%s) "
						"given for virtual service %s",
						value, service->name);
				free(service);
				return NULL;
			}
			service->ipvs_service.protocol = proto->p_proto;
		} else if (!xmlStrcmp(node->name,
					(const xmlChar *)"address")) {
			struct hostent *host;
			value = xmlNodeListGetString(node->doc,
					node->children,	1);
			if (!value) {
				log_printf(LOG_ERR, "No value given for "
						"address in virtual service %s",
						service->name);
				free(service);
				return NULL;
			}
			if (!(host = gethostbyname(value))
				|| host->h_addrtype != AF_INET
				|| host->h_length != sizeof(struct in_addr)) {
				log_printf(LOG_ERR, "Could not resolve "
						"address %s for virtual "
						"service %s", value,
						service->name);
				free(service);
				return NULL;
			}
		
			memcpy(&service->ipvs_service.addr,
					host->h_addr_list[0],
					sizeof(uint32_t));

		} else if (!xmlStrcmp(node->name,
		                     (const xmlChar *)"scheduler")) {
			value = xmlNodeListGetString(node->doc,
					node->children, 1);
			strncpy(service->ipvs_service.sched_name, value,
					IP_VS_SCHEDNAME_MAXLEN - 1);
			
		}

		node = node->next;
	}
	/* check service */
	if (!service->name || !strlen(service->name)) {
		log_printf(LOG_ERR, "No name given for virtual service");
		free(service);
		return NULL;
	}

	if (!service->ipvs_service.addr) {
		log_printf(LOG_ERR, "No address given for virtual service");
		free(service);
		return NULL;
	}

	if (!service->ipvs_service.protocol) {
		log_printf(LOG_ERR, "No protocol given for virtual service %s",
		             service->name);
		free(service);
		return NULL;
	}
	
	/* only look for port if the protocol is tcp or udp */
	if (service->ipvs_service.protocol == IPPROTO_TCP ||
			service->ipvs_service.protocol == IPPROTO_UDP) {
		long port;
		char *end;
		
		if (!tmpserv) {
			log_printf(LOG_ERR, "No port given for service %s",
			     	service->name);
				free(service);
			return NULL;
		}

		errno = 0;
		port = strtol(tmpserv, &end, 10);
		
		if (*end == '\0' && end != tmpserv) {
			/* it's a number */
			if (errno == ERANGE || port < 0 || port > 0xffff) {
				log_printf(LOG_ERR, "Invalid port (%s) given "
						"for service %s",
					     tmpserv, service->name);
				free(service);
				return NULL;
			}
			service->ipvs_service.port = htons(port);
		} else {
			
			if (!(serv = getservbyname(tmpserv, proto->p_name))) {
				log_printf(LOG_ERR, "Invalid port (%s) given "
						"for service %s",
						tmpserv, service->name);
				free(service);
				return NULL;
			}
			service->ipvs_service.port = serv->s_port;
		}
	} else {
		log_printf(LOG_WARN, "Service has a protocol of %s, which is "
				"not supported by LVS",
				service->ipvs_service.protocol);
		service->ipvs_service.port = 0;
	}

	return service;
}

static int configure(const char *configfile, int logging)
{
	xmlNodePtr lognode = NULL, node;
	xmlDocPtr doc;

	assert(configfile && strlen(configfile));

	doc = xmlParseFile(configfile);
	if (!doc) {
		log_printf(LOG_ERR, "Could not parse config file %s",
				configfile);
		return -1;
	}

	node = xmlDocGetRootElement(doc);
	if (!node) {
		log_printf(LOG_ERR, "Configuration file %s is empty",
		    configfile);
		xmlFreeDoc(doc);
		return -1;
	}

	if (xmlStrcmp(node->name, (const xmlChar *) "feedbackd-master")) {
		log_printf(LOG_ERR, "Configuration file %s has an incorrect "
				"root node (found %s, expected %s)",
				node->name, "feedbackd-master");
		xmlFreeDoc(doc);
		return -1;
	}

	node = node->children;
	while (node) {
		/* char *value; */
		
		if (xmlIsBlankNode(node)) {
			node = node->next;
			continue;
		}
		if (!xmlStrcmp(node->name, (const xmlChar*)"log")) {
			lognode = node;
		} else if (!xmlStrcmp(node->name,
			    (const xmlChar*)"scheduler")) {
			if (scheduler_configure(node)) {
				log_printf(LOG_ERR, "Scheduler configuration "
						"failed");
				xmlFreeDoc(doc);
				return -1;
			}
		} else if (!xmlStrcmp(node->name,
			    (const xmlChar*)"virtual-service")) {
			struct virtual_service *service =
				parse_virtual_service(node);

			if (!service) {
				log_printf(LOG_ERR, "virtual service "
						"configuration failed");
				xmlFreeDoc(doc);
				return -1;
			}

			if (list_add(&virtual_services, service)) {
				log_printf(LOG_ERR, "Error adding to service "
						"list");
				xmlFreeDoc(doc);
				return -1;
			}
		}
		node = node->next;
	}

	if (logging && log_configure(lognode)) {
		log_printf(LOG_ERR, "Could not configure logging");
		xmlFreeDoc(doc);
		return -1;
	}

	xmlFreeDoc(doc);
	return 0;

}
void dump_config()
{
	log_printf(LOG_DEBUG, "-- Configuration dump - virtual services:");
	listitem *i;

	for (i = virtual_services.head; i; i = i->next) {
		struct virtual_service *service =
			(struct virtual_service *)i->item;
		log_printf(LOG_DEBUG,
				"Service %s: ("
				"addr: %d.%d.%d.%d, protocol: %d, "
				"port: %d, sched: %s)",
				service->name,
				NIPQUAD(service->ipvs_service.addr),
				service->ipvs_service.protocol,
				ntohs(service->ipvs_service.port),
				service->ipvs_service.sched_name);
	}
}


void print_usage(const char *progname)
{
	fprintf(stderr,
"Usage: %s [-c|--conffile FILE] [-n|--dryrun]\n"
"           [-d|--dont-fork] [-V|--version]\n"
"Options:\n"
"    --conffile FILE:  use FILE to source configuration data\n"
"                      (default: %s)\n"
"\n"
"    --dryrun:         don't change ipvs tables (ipvs tables are still used \n"
"                      to check the number of current connections).\n"
"\n"
"    --dont-fork:      don't run in the background.\n"
"\n"
"    --version:        print version information and exit\n"
, progname, FEEDBACKD_CONF);
}


